#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include "matrix.h"

#define MAX_VALUE 100

/*
 * Write n zero values to the file
 * fd : file descriptor
 * n  : elements to write on the file
 * On success, 0 is returned. On error, -1 is returned.
 */
static int gen_zero(int fd, int n);

/*
 * Write n values to the file
 * 0,1,...,(MAX_VALUE-1),0,1,...
 * fd : file descriptor
 * n  : elements to write on the file
 * On success, 0 is returned. On error, -1 is returned.
 */
static int gen_incr(int fd, int n);

/*
 * Write n random values to the file
 * fd : file descriptor
 * n  : elements to write on the file 
 * On success, 0 is returned. On error, -1 is returned.
 */
static int gen_rand(int fd, int n);

/*
 * Set the file offset to the specified matrix element
 * fd : file descriptor
 * cols : number of columns
 * i : row index
 * j : column index
 * On success, 0 is returned. On error, -1 is returned.
 */
static int set_fd_offset(int fd, int cols, int i, int j);

/*
 * Lock the specified region of the file.
 * In test mode (cmd == F_GETLK), the function checks if the
 * lock is available and prints it to the standard output stream.
 * In normal mode, the functions reads or writes data and unlock
 * the specified region.
 * fd : file descriptor
 * cmd : lock command (F_GETLK for testing mode, F_SETLKW otherwise)
 * lock : lock structure
 * values : pointer to the buffer where the function reads or writes data.
 *          The buffer size must be equal to lock->l_len bytes
 * On success, 0 is returned. On error, -1 is returned.
 */
static int lock_unlock(int fd, int cmd, struct flock *lock, void *values);

#ifdef DEBUG
# define PAUSE(str) {\
   printf("%s.\nPress enter to continue.\n", str);\
   (void)getchar();\
 }
# define PRINT(str) printf("%s.\n", str)
#else
# define PAUSE(str)
# define PRINT(str)
#endif /* !DEBUG */

int mtx_gen(int fd, int rows, int cols, int mode) {
  off_t size;
  struct flock lock;
  struct matrix_hd mh;
  int ret;
  /* Calculate requested size */
  /* ... */

  /* Truncate the file to the specified length */
  /* ... */

  /* Take the exclusive lock to all the file */
  /* ... */

  /* Write the matrix header */
  ret = write(fd, &mh, sizeof(struct matrix_hd));
  if (ret > 0) {
	/* Write the matrix data */
	switch(mode) {
	case GEN_INCR:
	  ret = gen_incr(fd, rows*cols);
	  break;
	case GEN_RAND:
	  ret = gen_rand(fd, rows*cols);
	  break;
	default:
	  ret = gen_zero(fd, rows*cols);
	}
  }
  /* Release the lock */
  /* ... */

  return(ret);
}

int mtx_get_header(int fd, struct matrix_hd *mh) {
  struct flock lock;
  int ret;
  /* Take the shared lock */
  /* ... */

  /* Read the header */
  ret = read(fd, mh, lock.l_len);
  /* Release the lock */
  /* ... */

  printf("Process %d loads the matrix %dx%d.\n", (int)getpid(), mh->rows, mh->cols);
  return((ret == lock.l_len) ? 0 : -1);
}

int mtx_get_value(int fd, struct matrix_hd *mh, int i, int j, int *value, int test) {
  struct flock lock;
  /* Reposition the file offset */
  if (set_fd_offset(fd, mh->cols, i, j) == -1) {
	return(-1);
  }
  /* Take the shared lock */
  /* ... use lock_unlock() function */
}

int mtx_set_value(int fd, struct matrix_hd *mh, int i, int j, int value, int test) {
  struct flock lock;
  /* Reposition the file offset */
  if (set_fd_offset(fd, mh->cols, i, j) == -1) {
	return(-1);
  }
  /* Take the exclusive lock */
  /* ... use lock_unlock() function */
}

int mtx_get_row(int fd, struct matrix_hd *mh, int i, int *values, int test) {
  struct flock lock;
  /* Reposition the file offset */
  if (set_fd_offset(fd, mh->cols, i, 0) == -1) {
	return(-1);
  }
  /* Take the shared lock */
  /* ... use lock_unlock() function */
}

int mtx_set_row(int fd, struct matrix_hd *mh, int i, int *values, int test) {
  struct flock lock;
  /* Reposition the file offset */
  if (set_fd_offset(fd, mh->cols, i, 0) == -1) {
	return(-1);
  }
  /* Take the exclusive lock */
  /* ... use lock_unlock() function */
}

static int gen_zero(int fd, int n) {
  int value;
  value = 0;
  while (n--) {
	if (write(fd, &value, sizeof(int)) < 0) {
	  perror("write");
	  return(-1);
	}
  }
  return(0);
}

static int gen_incr(int fd, int n) {
  int value;
  value = 0;
  while (n--) {
	if (write(fd, &value, sizeof(int)) < 0) {
	  perror("write");
	  return(-1);
	}
	value++;
	if (value >= MAX_VALUE) {
	  value = 0;
	}
  }
  return(0);
}

static int gen_rand(int fd, int n) {
  int value;
  /* Set random seed */
  srand((unsigned int)getpid());
  while(n--) {
	value = rand() % MAX_VALUE;
	if (write(fd, &value, sizeof(int)) < 0) {
	  perror("write");
	  return(-1);
	}
  }
  return(0);
}

static int set_fd_offset(int fd, int cols, int i, int j) {
  off_t offset;
  offset = sizeof(struct matrix_hd) + (i*cols + j)*sizeof(int);
  /* Reposition the file offset */
  if (lseek(fd, offset, SEEK_SET) == (off_t)-1) {
	perror("lseek");
	return(-1);
  }
  return(0);
}

static int lock_unlock(int fd, int cmd, struct flock *lock, void *values) {
  int ret;
  if (fcntl(fd, cmd, lock) == -1) {
	perror("fcntl lock");
	return(-1);
  }
  if (cmd == F_GETLK) {
	/* Test mode */
	if (lock->l_type == F_UNLCK) {
	  printf("Requested lock is available.\n");
	}
	else {
	  printf("Requested lock is unavailable. Process %d holds an incompatible lock.\n",
			 (int)lock->l_pid);
	}
	return(0);
  }
  /* Normal mode */
  PAUSE("Took the lock");
  /* Write or read data */
  ret = (lock->l_type == F_RDLCK) ? 
	read(fd, values, lock->l_len) : write(fd, values, lock->l_len);
  /* Release the lock */
  lock->l_type = F_UNLCK;
  if (fcntl(fd, F_SETLK, lock) == -1) {
	perror("fcntl unlock");
	return(-1);
  }
  PRINT("Lock released");
  return((ret == lock->l_len) ? 0 : -1);
}
