#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "matrix.h"

/* Options */
struct options {
  int row;
  int col;
  int test;
  char *file;
};

/* Parse command-line options */
static void parse_args(int argc, char **argv, struct options *opt);

int main(int argc, char **argv) {
  int fd, *values, x;
  struct options opt;
  struct matrix_hd mh;
  /* Parse command-line options */
  parse_args(argc, argv, &opt);
  /* Open the matrix file */
  if ((fd = open(opt.file, O_RDONLY)) == -1) {
	perror(opt.file);
	return(EXIT_FAILURE);
  }
  /* Retrieve matrix header */
  if (mtx_get_header(fd, &mh) == -1) {
	close(fd);
	return(EXIT_FAILURE);
  }
  /* Check if the row-id is valid */
  if (opt.row >= mh.rows || opt.row < 0) {
	printf("Error: Invalid row.\n");
	close(fd);
	return(EXIT_FAILURE);
  }
  if (opt.col >= 0) {
	/* Check if the column-id is valid */
	if (opt.col >= mh.cols || opt.col < 0) {
	  printf("Error: Invalid column.\n");
	  close(fd);
	  return(EXIT_FAILURE);
	}
	if (mtx_get_value(fd, &mh, opt.row, opt.col, &x, opt.test) == -1) {
	  close(fd);
	  return(EXIT_FAILURE);
	}
	if (!opt.test) {
	  printf("M[%d][%d] = %d\n", opt.row, opt.col, x);
	}
  }
  else {
	if (!opt.test) {
	  /* Allocate memory */
	  if (!(values = malloc(sizeof(int)*mh.cols))) {
		perror("malloc");
		close(fd);
		return(EXIT_FAILURE);
	  }
	  if (mtx_get_row(fd, &mh, opt.row, values, 0) == -1) {
		close(fd);
		return(EXIT_FAILURE);
	  }
	  printf("M[%d][*] =", opt.row);
	  for (x = 0; x < mh.cols; x++) {
		printf(" %2d", values[x]);
	  }
	  printf("\n");
	  /* Free memory */
	  free(values);
	}
	else {
	  /* Test mode */
	  if (mtx_get_row(fd, &mh, opt.row, NULL, 1) == -1) {
		close(fd);
		return(EXIT_FAILURE);
	  }
	}
  }
  /* Close the file */
  if (close(fd) == -1) {
	perror("close");
	return(EXIT_FAILURE);
  }
  return(EXIT_SUCCESS);
}

/* print help */
#define PRINT_HELP(prg)	\
  printf("Usage: %s -f filename -r row_id [-c col_id] [-t] [-h]\n", prg)

/* print help and exit with failure status */
#define ERROR(prg,str) {\
  printf("Error: %s.\n", str);\
  PRINT_HELP(prg);\
  exit(EXIT_FAILURE);\
}

static void parse_args(int argc, char **argv, struct options *opt) {
  int c;
  /* Reset options */
  opt->file = NULL;
  opt->row = opt->col = -1;
  opt->test = 0;
  /* Parse command-line options */
  while ((c = getopt(argc, argv, "f:r:c:th")) != -1) {
	switch(c) {
	case 'r':
	  /* Row-id */
	  if (opt->row != -1) {
		ERROR(argv[0], "You have specified the matrix row id twice");
	  }
	  opt->row = atoi(optarg);
	  break;
	case 'c':
	  /* Column-id */
	  if (opt->col != -1) {
		ERROR(argv[0], "You have specified the matrix column id twice");
	  }
	  opt->col = atoi(optarg);
	  break;
	case 'f':
	  /* Filename */
	  if (opt->file) {
		ERROR(argv[0], "You can specify only one filename");
	  }
	  opt->file = optarg;
	  break;
	case 't':
	  /* Test mode */
	  opt->test = 1;
	  break;
	case 'h':
	  PRINT_HELP(argv[0]);
	  exit(EXIT_SUCCESS);
	default:
	  ERROR(argv[0], "Invalid argument");
	}
  }
  /* Check arguments */
  if (!opt->file || opt->row == -1) {
	PRINT_HELP(argv[0]);
	exit(EXIT_FAILURE);
  }
}
