#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>
#include "mmap_matrix.h"
#include "test.h"

#define OP_GEN   1
#define OP_TEST  2
#define OP_PRINT 3

/* Matrix arguments */
struct margs {
  int rows;   /* Rows */
  int cols;   /* Columns */
  char *file; /* File name */
  int op;     /* Operation: OP_GEN, OP_TEST, OP_PRINT */
  short mode; /* Generation mode: GEN_ZERO, GEN_RAND, GEN_INCR */
};

/* Print help */
#define PRINT_HELP(prg) printf("Usage:\n"\
						  "\tGenerate a matrix (rows*cols) into 'filename'\n"\
						  "\t\t%s -g <filename> -r <rows> -c <cols> -m <r|i|z>\n"\
						  "\tPrint the matrix from 'filename'\n"\
						  "\t\t%s -p <filename>\n"\
						  "\tCheck if the matrix from 'filename' is correct\n"\
						  "\t\t%s -t <filename>\n\n", prg, prg, prg)

/* Print error, help and exit */
#define ERROR(str) {\
  printf("Error: %s.\n", str);\
  PRINT_HELP(argv[0]);\
  exit(EXIT_FAILURE);\
}

static void parse_args(int argc, char **argv, struct margs *m);
static int matrix_generate(char *filename, int rows, int cols, short mode);
static int matrix_check(const char *filename);
static int matrix_print(const char *filename);

/* Main */
int main(int argc, char **argv) {
  struct margs m;
  int ret;
  parse_args(argc, argv, &m);
  switch (m.op) {
  case OP_TEST:
	ret = matrix_check(m.file);
	break;
  case OP_PRINT:
	ret = matrix_print(m.file);
	break;
  case OP_GEN:
	if (m.rows <= 0 || m.cols <= 0) {
	  ERROR("You must specify the rows and/or the columns of the matrix");
	}
	ret = matrix_generate(m.file, m.rows, m.cols, m.mode);
	break;
  default:
	ERROR("You must specify one operation");
  }
  return(ret ? EXIT_FAILURE : EXIT_SUCCESS);
}

static int matrix_generate(char *filename, int rows, int cols, short mode) {
  int fd;
  /* Create the file */
  if ((fd = open(filename, O_RDWR|O_CREAT, 0640)) == -1) {
	perror(filename);
	return(-1);
  }
  /* Generate the matrix */
  mode = matrix_gen(fd, rows, cols, mode);
  if (close(fd)) {
	perror(filename);
  }
  return(mode);
}

#define TESTS 1000

int matrix_check(const char *filename) {
  int fd;
  /* Open the file */
  if ((fd = open(filename, O_RDWR)) == -1) {
	perror(filename);
	return(-1);
  }
  /* Set the seed */
  srand((unsigned int) time(NULL));
  /* Run the tests */
  if ((test1(fd, TESTS) == -1) ||
	  (test2(fd, TESTS) == -1)) {
	close(fd);
	return(-1);
  }
  printf("All tests (%d) are passed successfully\n", TESTS*2);
  if (close(fd)) {
	perror(filename);
  }
  return(0);
}

int matrix_print(const char *filename) {
  MATRIX m;
  int fd, i, j;
  /* Open the file */
  if ((fd = open(filename, O_RDWR)) == -1) {
	perror(filename);
	return(-1);
  }
  /* Load the matrix */
  if (matrix_load(fd, &m) == -1) {
	close(fd);
	return(-1);
  }
  /* Print the matrix */
  printf("Rows: %d - Columns: %d\n", ROWS(&m), COLS(&m));
  for (i = 0; i < ROWS(&m); i++) {
	printf("%02d:", i);
	for (j = 0; j < COLS(&m); j++) {
	  printf(" %3d", M(&m, i, j));
	}
	printf("\n");
  }
  /* Unmap the matrix */
  if (matrix_store(&m) == -1) {
	close(fd);
	return(-1);
  }
  if (close(fd)) {
	perror(filename);
  }
  return(0);
}

static void parse_args(int argc, char **argv, struct margs *m) {
  int c;
  m->rows = m->cols = m->op = 0;
  m->mode = 0;
  /* Parse arguments */
  while ((c = getopt(argc, argv, "t:p:g:m:r:c:h")) != -1) {
	switch (c) {
	case 't':
	  /* Test operation */
	  if (m->op) {
		ERROR("You can specify only one operation");
	  }
	  m->op = OP_TEST;
	  m->file = optarg;
	  break;
	case 'p':
	  /* Print operation */
	  if (m->op) {
		ERROR("You can specify only one operation");
	  }
	  m->op = OP_PRINT;
	  m->file = optarg;
	  break;
	case 'g':
	  /* Generate operation */
	  if (m->op) {
		ERROR("You can specify only one operation");
	  }
	  m->op = OP_GEN;
	  m->file = optarg;
	  break;
	case 'r':
	  /* Rows */
	  m->rows = atoi(optarg);
	  break;
	case 'c':
	  /* Columns */
	  m->cols = atoi(optarg);
	  break;
	case 'm':
	  /* Generation mode */
	  if (m->mode) {
		ERROR("You can specify only one generation mode");
	  }
	  switch(*optarg) {
	  case 'r':
		m->mode = GEN_RAND;
		break;
	  case 'i':
		m->mode = GEN_INCR;
		break;
	  default:
		m->mode = GEN_ZERO;
	  }
	  break;
	case 'h':
	  PRINT_HELP(argv[0]);
	  exit(EXIT_SUCCESS);
	default:
	  printf("getopt returned character code 0%o\n", c);
	  exit(EXIT_FAILURE);
	}
  }
}
