// Server program to demonstrate usage of FIFOs (named-pipes) 
// ----------------------------------------------------------
// LSO1 AA2005-2006 - M. Verola
// ----------------------------
// This server program "selectServer2" creates a certain number 
// of FIFOs in pairs (read+write), and monitors in an infinite 
// loop all the read-FIFOs for incoming messages. 
// The number of FIFOs shall be specified as the first command 
// line argument.
// To quit the program, type Ctrl+C.
// To test this server program, the client program "selectClient"
// shall be used.
//
// When a message is detected on any read-FIFO, it is read,
// processed (i.e. converted into upper case) and then sent back
// to the client program, that was the message originator.
// To be more precise, the returned message is sent through
// the write-FIFO associated in pair with the read-FIFO that
// contained the message.
//
// The default naming convention for the read/write FIFOs is:
// readFIFO.<n>, writeFIFO.<n>, where <n> is a number between
// 0 and N-1, with N equal to the first command line argument 
// of this program.
// The base name of the FIFOs can be changed (see include file
// selectCommon.h).
//
// NOTE: this program is just an example for the usage of some
//       system calls, in particular select(), signal() and FIFO-related
//       system calls. The various messages on stdout are for
//       tracing purposes and should be removed when used
//       for real services. The error management is very simple
//       and not exhaustive.
//
#include "selectCommon.h"

#define MYAPP "[selectServer]"

// Global variables
int fifos;
int *rfds;
int *listFifoReady;

// Create the <basename>.<n> FIFOs.
int createFifos(int fifos, const char *basename) {
  char fifoname[MAXFILENAME];
  int i;

  /* Create <fifos> FIFOs for reading and <fifos> FIFOs for writing */
  for (i = 0; i < fifos; i++) {
    snprintf(fifoname, MAXFILENAME, "%s.%d", basename, i);
    printf("%s Creating the FIFO \"%s\"\n", MYAPP, fifoname);
    if ((mkfifo(fifoname, 0666) == -1) && (errno != EEXIST)) {
      perror("mkfifo");
      return(-1);
    }    
  }
  
  return(0);
}

// Open every readFIFO.<n> for reading messages coming 
// from clients and compute the max value of the FIFO desciptor,
// to be used for select().
// Actually every readFIFO.<n> in NOT opened just as 
// O_RDONLY, but as O_RDWR. This is a trick, to prevent the
// following unconvenient situation: after the client has closed the 
// write-end FIFO descriptor, the select() returns always FIFO
// descriptor ready (even though no bytes are inside) and the read()
// returns immediately with 0 bytes read.
// Hence, to avoid this behaviour, the server itself opens the readFIFO.<n>
// both for reading and for writing, so that there is always a process
// that keeps the write-end of the readFIFO.<n> opened. 
// IN  : fifos, rbname
// OUT : rfds, maxFdP
int openReadFifos(int fifos, char *rbname, int *rfds, int *maxFdP) {
  char fifoname[MAXFILENAME];
  int i, max;
  max = 0;

  // Open <fifos> FIFO for reading, but, to avoid problems when
  // the client closes the write-end of the read-FIFO, we have
  // to open these read-FIFOs in O_RDWR mode!
  // This is a trick, to be kept in mind!!! 
  for (i = 0; i < fifos; i++) {
    snprintf(fifoname, MAXFILENAME, "%s.%d", rbname, i);
    if ((rfds[i] = open(fifoname, O_RDWR)) == -1) {
      perror("open");
      return(-1);
    }
    printf("%s FIFO \"%s\" (fd=%d) opened for reading (actually with O_RDWR!)\n", MYAPP, fifoname, rfds[i]);
    max = (rfds[i] > max) ? rfds[i] : max;
  }
  *maxFdP = max;

  return(0);
}

void closeReadFifos(int fifos, int *rfds) {
  while (fifos--) {
    printf("%s Closing file descriptor %d\n", MYAPP, *rfds);
    close(*rfds++); 
  }
}

void unlinkFifos(int fifos, const char *basename) {
  int i;
  char fifoname[MAXFILENAME];
  for (i = 0; i < fifos; i++) {
    snprintf(fifoname, MAXFILENAME, "%s.%d", basename, i);
    printf("%s Removing the FIFO \"%s\"\n", MYAPP, fifoname);   
    unlink(fifoname);  
  }
}

// Given a descriptor value, searches the index array of the descriptor
// as stored in the rfds[] array
// IN  : fifos, rfds, rfd
// OUT : index
int getReadFifoIndex(int fifos, int *rfds, int rfd, int *index) {
  int i;

  for (i = 0; i < fifos; i++) {
    if (rfds[i] == rfd) {
	  // item found
	  *index = i;
	  printf("%s FIFO index corresponding to fd=%d found => %s.%d!\n", MYAPP, rfd, RFIFO_BASE_NAME, i);
	  return(0);
    }
  }
  // item not found
  printf("%s FIFO index corresponding to fd=%d NOT found!\n", MYAPP, rfd);
  return(-1);
}

// IN  : fifos, maxFd, rfds
// OUT : numFifoReady, listFifoReady
int wakeupOnMsg(int fifos, int maxFd, int *rfds, int *numFifoReady, int *listFifoReady) {
  int nFdReady, i, n;
  fd_set readFds;

  // Init every time the read-fifo fd set for select() 
  // Monitor only the read-fifos
  FD_ZERO(&readFds); // Reset readFds set
  for (i = 0; i < fifos; i++) {
    // Set the bits in readFds corresponding to the read-pipe descriptor numbers 
    FD_SET(rfds[i], &readFds);
  } 
 
  printf("%s Waiting in select() on %d FIFO descriptors ...\n", MYAPP, fifos);
 
  // Waiting for a msg from any of the read-fifos
  n = select(maxFd+1, &readFds, NULL, NULL, NULL); 
  if (n < 0) {
    perror("select() failed!");
    if (errno == EINTR)
      return(-2); // Interrupt received, so select() has exited
    else
      return(-1); // Error in select
  }
  printf("%s select() returned %d!\n\n",MYAPP,n);

  // Count the read-fifos that are ready (i.e. with some data to be read)
  // and fill an array with their rfds
  for (i = 0, nFdReady = 0; i < fifos; i++) {
    if (FD_ISSET(rfds[i], &readFds)) {
      printf("%s Read-FIFO %s.%d (fd=%d) is ready!\n", MYAPP, RFIFO_BASE_NAME, i, rfds[i]);
      listFifoReady[nFdReady++] = rfds[i]; // Store this fd to be returned to caller
    }
  }
  *numFifoReady = nFdReady;
 
  return(0);
}

int replyToClient(char *buf, int fdReady) {
  char fifoname[MAXFILENAME];
  int len, index, wfd;

  // Get the array index corresponding to this read-pipe fdReady,
  // so that we can open the proper write-FIFO, that is the one
  // with the same suffix of the read-FIFO, from where we read the msg:
  // read msg from readFIFO.x -> write processed msg to writeFIFO.x
  if (getReadFifoIndex(fifos, rfds, fdReady, &index) == -1) {
	return(-1);
  }

  // Build the name of the write-fifo and open it for writing
  snprintf(fifoname, MAXFILENAME, "%s.%d", WFIFO_BASE_NAME, index);
  printf("%s Opening the FIFO \"%s\" for writing (might block)...\n", MYAPP, fifoname);
  // Open FIFO for writing
  if ((wfd = open(fifoname, O_WRONLY)) == -1) {
    perror("open ");
    return(-1);
  }
  printf("%s ... FIFO \"%s\" opened with fd=%d\n", MYAPP, fifoname, wfd);
  
  /* Write the processed msg back to the write-fifo */    
  printf("%s Writing to the FIFO \"%s\" (fd=%d)\n", MYAPP, fifoname, wfd);
  len = strlen(buf) + 1;
  if (write(wfd, buf, len) != len) {
	perror("write");
  }
  close(wfd);
  return(0);
}

// IN  : numread
// OUT : buf
void processMsg(int numread, char *buf) {
  /* Convert the string to upper case */
  while (numread--) {
	*buf = toupper(*buf);
	buf++;
  }
}

// This signal handler runs when SIGINT (Ctrl+C) is caught.
// This function shall clean up the server's resources.
void cleanUp(int sig) {
  printf("\n%s Signal %d caught!\n", MYAPP, sig);
  printf("%s Clean up the environment and exit!\n", MYAPP);
  closeReadFifos(fifos, rfds);
  unlinkFifos(fifos, RFIFO_BASE_NAME);
  unlinkFifos(fifos, WFIFO_BASE_NAME);
  // Free memory
  if (rfds) {
    free(rfds);
  }
  if (listFifoReady) {
    free(listFifoReady);
  }
  exit(EXIT_SUCCESS);
}

//==========================================================
// MAIN FUNCTION
//==========================================================
int main(int argc, char *argv[]) {
  int numread, maxRfd, nFifoReady;
  char buf[MAX_BUF_SIZE], bufold[MAX_BUF_SIZE];
  int i;

  if (argc != 2) {
    fprintf(stderr, "\nUsage\t%s <nfifos>", argv[0]);
    fprintf(stderr, "\n     \tCtrl-C to quit\n");
    exit(EXIT_FAILURE);
  }

  // Register the signal handler for the SIGINT and SIGTERM signal
  signal(SIGINT,cleanUp);
  signal(SIGTERM,cleanUp);

  // The first arg on command line is the number of fifo couples (r+w)
  //  to be created
  fifos = atoi(argv[1]);

  // Create the readFIFO.<n> and writeFIFO.<n> FIFOs.
  if (createFifos(fifos, RFIFO_BASE_NAME) == -1 ||
	  createFifos(fifos, WFIFO_BASE_NAME) == -1) {
	return(EXIT_FAILURE);
  }

  // Allocate an array of <fifos> items to store the readFIFO.<n>
  // descriptors.
  if (!(rfds = malloc(fifos*sizeof(int)))) {
    perror("malloc");
    return(EXIT_FAILURE);
  }
  // Open FIFOs
  if (openReadFifos(fifos, RFIFO_BASE_NAME, rfds, &maxRfd) == -1) {
	return(EXIT_FAILURE);
  }

  // Allocate an array to be used to store read-fifo fd that select()
  //  will find ready, that is with some bytes to be read
  if (!(listFifoReady = malloc(fifos*sizeof(int)))) {
    perror("malloc");
    return(EXIT_FAILURE);
  }

  // Loop
  for (;;) {

    // Wait on an incoming message
    if (wakeupOnMsg(fifos, maxRfd, rfds, // IN
					&nFifoReady, listFifoReady) < 0) { // OUT
	  return(EXIT_FAILURE);
	}
	
    // Read all the read-fifos that were found ready
    // and reply using the corresponding write-fifo
    for (i = 0; i < nFifoReady; i++) {
      // Read from the read-fifo
      buf[0]='\0';
      numread = read(listFifoReady[i], buf, MAX_BUF_SIZE);
      if (numread < 0) { // real error
		perror("read");
		return(EXIT_FAILURE);
      }
      if (!numread) {
		printf("%s ZERO BYTES read (probably the write-end of FIFO has been closed).\n", MYAPP);
		continue;
      }
	  
	  printf("%s Message read from the read-FIFO with fd=%d: <%s> (%d bytes)\n",
			 MYAPP, listFifoReady[i], buf, numread);

      // Save a copy of the original message
      strcpy(bufold, buf);

      // Process the received msg
      processMsg(numread, buf); 

      printf("%s Message <%s> has been converted into <%s>\n", MYAPP, bufold, buf);
      
	  // Reply to client
	  if (replyToClient(buf, listFifoReady[i]) == -1) {
		return(EXIT_FAILURE);
	  }
    }
  }
   
  exit(EXIT_SUCCESS);
}
